def audit(policy):
    # The following list is obtained from Policy Sentry via the following code.
    # This is done to avoid pulling in Policy Sentry and its requirements which adds ~50MB to this library.
    """
    from policy_sentry.shared.database import connect_db
    from policy_sentry.querying.actions import get_actions_with_access_level

    db_session = connect_db("bundled")
    permissions_management_actions = get_actions_with_access_level(
        db_session, "all", "Permissions management"
    )
    permissions_management_actions_normalized = [
        x.lower() for x in permissions_management_actions
    ]
    permissions_management_actions = permissions_management_actions_normalized
    """

    permissions_management_actions = [
        "acm-pca:createpermission",
        "acm-pca:deletepermission",
        "backup:deletebackupvaultaccesspolicy",
        "backup:putbackupvaultaccesspolicy",
        "chime:deletevoiceconnectorterminationcredentials",
        "chime:putvoiceconnectorterminationcredentials",
        "cloudformation:setstackpolicy",
        "cloudsearch:updateserviceaccesspolicies",
        "codebuild:deleteresourcepolicy",
        "codebuild:deletesourcecredentials",
        "codebuild:importsourcecredentials",
        "codebuild:putresourcepolicy",
        "codestar:associateteammember",
        "codestar:createproject",
        "codestar:deleteproject",
        "codestar:disassociateteammember",
        "codestar:updateteammember",
        "cognito-identity:createidentitypool",
        "cognito-identity:deleteidentities",
        "cognito-identity:deleteidentitypool",
        "cognito-identity:getid",
        "cognito-identity:mergedeveloperidentities",
        "cognito-identity:setidentitypoolroles",
        "cognito-identity:unlinkdeveloperidentity",
        "cognito-identity:unlinkidentity",
        "cognito-identity:updateidentitypool",
        "connect:getfederationtoken",
        "connect:getfederationtokens",
        "deeplens:associateserviceroletoaccount",
        "ds:createconditionalforwarder",
        "ds:createdirectory",
        "ds:createidentitypooldirectory",
        "ds:createmicrosoftad",
        "ds:createtrust",
        "ds:sharedirectory",
        "ec2:createnetworkinterfacepermission",
        "ec2:deletenetworkinterfacepermission",
        "ec2:modifysnapshotattribute",
        "ec2:modifyvpcendpointservicepermissions",
        "ec2:resetsnapshotattribute",
        "ecr:setrepositorypolicy",
        "elasticmapreduce:putblockpublicaccessconfiguration",
        "es:createelasticsearchdomain",
        "es:updateelasticsearchdomainconfig",
        "gamelift:requestuploadcredentials",
        "glacier:abortvaultlock",
        "glacier:completevaultlock",
        "glacier:deletevaultaccesspolicy",
        "glacier:initiatevaultlock",
        "glacier:setdataretrievalpolicy",
        "glacier:setvaultaccesspolicy",
        "glue:deleteresourcepolicy",
        "glue:putresourcepolicy",
        "greengrass:associateserviceroletoaccount",
        "health:describehealthservicestatusfororganization",
        "health:disablehealthserviceaccessfororganization",
        "health:enablehealthserviceaccessfororganization",
        "iam:addclientidtoopenidconnectprovider",
        "iam:addroletoinstanceprofile",
        "iam:addusertogroup",
        "iam:attachgrouppolicy",
        "iam:attachrolepolicy",
        "iam:attachuserpolicy",
        "iam:changepassword",
        "iam:createaccesskey",
        "iam:createaccountalias",
        "iam:creategroup",
        "iam:createinstanceprofile",
        "iam:createloginprofile",
        "iam:createopenidconnectprovider",
        "iam:createpolicy",
        "iam:createpolicyversion",
        "iam:createrole",
        "iam:createsamlprovider",
        "iam:createservicelinkedrole",
        "iam:createservicespecificcredential",
        "iam:createuser",
        "iam:createvirtualmfadevice",
        "iam:deactivatemfadevice",
        "iam:deleteaccesskey",
        "iam:deleteaccountalias",
        "iam:deleteaccountpasswordpolicy",
        "iam:deletegroup",
        "iam:deletegrouppolicy",
        "iam:deleteinstanceprofile",
        "iam:deleteloginprofile",
        "iam:deleteopenidconnectprovider",
        "iam:deletepolicy",
        "iam:deletepolicyversion",
        "iam:deleterole",
        "iam:deleterolepermissionsboundary",
        "iam:deleterolepolicy",
        "iam:deletesamlprovider",
        "iam:deletesshpublickey",
        "iam:deleteservercertificate",
        "iam:deleteservicelinkedrole",
        "iam:deleteservicespecificcredential",
        "iam:deletesigningcertificate",
        "iam:deleteuser",
        "iam:deleteuserpermissionsboundary",
        "iam:deleteuserpolicy",
        "iam:deletevirtualmfadevice",
        "iam:detachgrouppolicy",
        "iam:detachrolepolicy",
        "iam:detachuserpolicy",
        "iam:enablemfadevice",
        "iam:passrole",
        "iam:putgrouppolicy",
        "iam:putrolepermissionsboundary",
        "iam:putrolepolicy",
        "iam:putuserpermissionsboundary",
        "iam:putuserpolicy",
        "iam:removeclientidfromopenidconnectprovider",
        "iam:removerolefrominstanceprofile",
        "iam:removeuserfromgroup",
        "iam:resetservicespecificcredential",
        "iam:resyncmfadevice",
        "iam:setdefaultpolicyversion",
        "iam:setsecuritytokenservicepreferences",
        "iam:updateaccesskey",
        "iam:updateaccountpasswordpolicy",
        "iam:updateassumerolepolicy",
        "iam:updategroup",
        "iam:updateloginprofile",
        "iam:updateopenidconnectproviderthumbprint",
        "iam:updaterole",
        "iam:updateroledescription",
        "iam:updatesamlprovider",
        "iam:updatesshpublickey",
        "iam:updateservercertificate",
        "iam:updateservicespecificcredential",
        "iam:updatesigningcertificate",
        "iam:updateuser",
        "iam:uploadsshpublickey",
        "iam:uploadservercertificate",
        "iam:uploadsigningcertificate",
        "imagebuilder:getcomponentpolicy",
        "imagebuilder:putcomponentpolicy",
        "imagebuilder:putimagepolicy",
        "imagebuilder:putimagerecipepolicy",
        "iot:attachpolicy",
        "iot:attachprincipalpolicy",
        "iot:detachpolicy",
        "iot:detachprincipalpolicy",
        "iot:setdefaultauthorizer",
        "iot:setdefaultpolicyversion",
        "iotsitewise:createaccesspolicy",
        "iotsitewise:deleteaccesspolicy",
        "iotsitewise:listaccesspolicies",
        "iotsitewise:updateaccesspolicy",
        "kms:creategrant",
        "kms:createkey",
        "kms:putkeypolicy",
        "kms:retiregrant",
        "kms:revokegrant",
        "lakeformation:batchgrantpermissions",
        "lakeformation:batchrevokepermissions",
        "lakeformation:grantpermissions",
        "lakeformation:putdatalakesettings",
        "lakeformation:revokepermissions",
        "lambda:addlayerversionpermission",
        "lambda:addpermission",
        "lambda:disablereplication",
        "lambda:enablereplication",
        "lambda:removelayerversionpermission",
        "lambda:removepermission",
        "license-manager:updateservicesettings",
        "lightsail:getinstanceaccessdetails",
        "lightsail:getrelationaldatabasemasteruserpassword",
        "logs:deleteresourcepolicy",
        "logs:putresourcepolicy",
        "mediapackage:rotateingestendpointcredentials",
        "mediastore:deletecontainerpolicy",
        "mediastore:putcontainerpolicy",
        "opsworks:setpermission",
        "opsworks:updateuserprofile",
        "ram:acceptresourceshareinvitation",
        "ram:associateresourceshare",
        "ram:createresourceshare",
        "ram:deleteresourceshare",
        "ram:disassociateresourceshare",
        "ram:enablesharingwithawsorganization",
        "ram:rejectresourceshareinvitation",
        "ram:updateresourceshare",
        "rds:authorizedbsecuritygroupingress",
        "rds-db:connect",
        "redshift:authorizesnapshotaccess",
        "redshift:createclusteruser",
        "redshift:createsnapshotcopygrant",
        "redshift:getclustercredentials",
        "redshift:joingroup",
        "redshift:modifyclusteriamroles",
        "redshift:revokesnapshotaccess",
        "s3:bypassgovernanceretention",
        "s3:deleteaccesspointpolicy",
        "s3:deletebucketpolicy",
        "s3:objectowneroverridetobucketowner",
        "s3:putaccesspointpolicy",
        "s3:putaccountpublicaccessblock",
        "s3:putbucketacl",
        "s3:putbucketpolicy",
        "s3:putbucketpublicaccessblock",
        "s3:putobjectacl",
        "s3:putobjectversionacl",
        "secretsmanager:deleteresourcepolicy",
        "secretsmanager:putresourcepolicy",
        "sns:addpermission",
        "sns:createtopic",
        "sns:removepermission",
        "sns:settopicattributes",
        "sqs:addpermission",
        "sqs:createqueue",
        "sqs:removepermission",
        "sqs:setqueueattributes",
        "ssm:modifydocumentpermission",
        "sso:associatedirectory",
        "sso:associateprofile",
        "sso:createapplicationinstance",
        "sso:createapplicationinstancecertificate",
        "sso:createpermissionset",
        "sso:createprofile",
        "sso:createtrust",
        "sso:deleteapplicationinstance",
        "sso:deleteapplicationinstancecertificate",
        "sso:deletepermissionset",
        "sso:deletepermissionspolicy",
        "sso:deleteprofile",
        "sso:disassociatedirectory",
        "sso:disassociateprofile",
        "sso:importapplicationinstanceserviceprovidermetadata",
        "sso:putpermissionspolicy",
        "sso:startsso",
        "sso:updateapplicationinstanceactivecertificate",
        "sso:updateapplicationinstancedisplaydata",
        "sso:updateapplicationinstanceresponseconfiguration",
        "sso:updateapplicationinstanceresponseschemaconfiguration",
        "sso:updateapplicationinstancesecurityconfiguration",
        "sso:updateapplicationinstanceserviceproviderconfiguration",
        "sso:updateapplicationinstancestatus",
        "sso:updatedirectoryassociation",
        "sso:updatepermissionset",
        "sso:updateprofile",
        "sso:updatessoconfiguration",
        "sso:updatetrust",
        "sso-directory:addmembertogroup",
        "sso-directory:createalias",
        "sso-directory:creategroup",
        "sso-directory:createuser",
        "sso-directory:deletegroup",
        "sso-directory:deleteuser",
        "sso-directory:disableuser",
        "sso-directory:enableuser",
        "sso-directory:removememberfromgroup",
        "sso-directory:updategroup",
        "sso-directory:updatepassword",
        "sso-directory:updateuser",
        "sso-directory:verifyemail",
        "storagegateway:deletechapcredentials",
        "storagegateway:setlocalconsolepassword",
        "storagegateway:setsmbguestpassword",
        "storagegateway:updatechapcredentials",
        "waf:deletepermissionpolicy",
        "waf:getchangetoken",
        "waf:putpermissionpolicy",
        "waf-regional:deletepermissionpolicy",
        "waf-regional:getchangetoken",
        "waf-regional:putpermissionpolicy",
        "wafv2:createwebacl",
        "wafv2:deletewebacl",
        "wafv2:updatewebacl",
        "worklink:updatedevicepolicyconfiguration",
        "workmail:resetpassword",
        "workmail:resetuserpassword",
        "xray:putencryptionconfig",
    ]

    actions = policy.get_allowed_actions()

    permissions_management_actions_in_policy = []
    for action in actions:
        if action in permissions_management_actions:
            permissions_management_actions_in_policy.append(action)
    if len(permissions_management_actions_in_policy) > 0:
        policy.add_finding(
            "PERMISSIONS_MANAGEMENT_ACTIONS",
            location={"actions": permissions_management_actions_in_policy},
        )



"""
For AWS policies using conditionals, checking a single valued condition key with a check
designed for multi-value condition keys results in "overly permissive policies"
https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_single-vs-multi-valued-condition-keys.html
"""
import re
from parliament import Policy
from parliament.misc import make_list

def audit(policy: Policy) -> None:
    global_single_valued_condition_keys = [
        "aws:CalledViaFirst",
        "aws:CalledViaLast",
        "aws:CurrentTime",
        "aws:EpochTime",
        "aws:FederatedProvider",
        "aws:MultiFactorAuthAge",
        "aws:MultiFactorAuthPresent",
        "aws:PrincipalAccount",
        "aws:PrincipalArn",
        "aws:PrincipalIsAWSService",
        "aws:PrincipalOrgID",
        "aws:PrincipalServiceName",
        "aws:PrincipalTag",
        "aws:PrincipalType",
        "aws:referer",
        "aws:RequestedRegion",
        "aws:RequestTag/*",
        "aws:ResourceTag/*",
        "aws:SecureTransport",
        "aws:SourceAccount",
        "aws:SourceArn",
        "aws:SourceIdentity",
        "aws:SourceIp",
        "aws:SourceVpc",
        "aws:SourceVpce",
        "aws:TokenIssueTime",
        "aws:UserAgent",
        "aws:userid",
        "aws:username",
        "aws:ViaAWSService",
        "aws:VpcSourceIp",
    ]

    for stmt in policy.statements:
        if "Condition" not in stmt.stmt:
            return

        conditions = stmt.stmt["Condition"]
        for condition in conditions:
            # The operator is the first element (ex. `StringLike`) and the condition_block follows it
            operator = condition[0]
            condition_block = condition[1]
            if re.match(r"^For(All|Any)Values:", operator):
                keys = list(k for k,_v in condition_block)
                if any(any(re.match(k, key) for k in global_single_valued_condition_keys) for key in keys):
                    policy.add_finding(
                        "SINGLE_VALUE_CONDITION_TOO_PERMISSIVE",
                        detail='Checking a single value conditional key against a set of values results in overly permissive policies.',
                    )



""" 
For AWS resource policies, check whether they use discouraged constructions.
See the [AWS Policy Troubleshooting Guide](https://docs.aws.amazon.com/IAM/latest/UserGuide/troubleshoot_policies.html).

AWS documentation discourages the use of NotPrincipal, NotAction and
NotResource, particularly with Allow. These constructs, by default, grant 
permissions, then Deny the ones explicitly listed. Instead, use an explicit 
Resource, Action or Principal in your Allow list.
"""

from typing import Iterable

import jsoncfg

from parliament import Policy


def get_stmts(policy: Policy) -> Iterable:
    if "jsoncfg.config_classes.ConfigJSONObject" in str(
        type(policy.policy_json.Statement)
    ):
        return [policy.policy_json.Statement]
    elif "jsoncfg.config_classes.ConfigJSONArray" in str(
        type(policy.policy_json.Statement)
    ):
        return policy.policy_json.Statement


def audit(policy: Policy) -> None:
    for stmt in get_stmts(policy):
        if stmt.Effect.value == "Allow" and jsoncfg.node_exists(stmt["NotPrincipal"]):
            # See https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_notprincipal.html#specifying-notprincipal-allow
            policy.add_finding(
                "NOTPRINCIPAL_WITH_ALLOW",
                location=("NotPrincipal", stmt["NotPrincipal"]),
            )
        elif stmt.Effect.value == "Allow" and jsoncfg.node_exists(stmt["NotResource"]):
            # See https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_notresource.html#notresource-element-combinations
            policy.add_finding(
                "NOTRESOURCE_WITH_ALLOW",
                location=("NotResource", stmt["NotResource"]),
            )



import jsoncfg


def make_list(v):
    """
    If the object is not a list already, it converts it to one
    Examples:
    [1, 2, 3] -> [1, 2, 3]
    [1] -> [1]
    1 -> [1]
    """
    if not jsoncfg.node_is_array(v):
        if jsoncfg.node_is_scalar(v):
            location = jsoncfg.node_location(v)
            line = location.line
            column = location.column
        elif jsoncfg.node_exists(v):
            line = v.line
            column = v.column
        else:
            return []

        a = jsoncfg.config_classes.ConfigJSONArray(line, column)
        a._append(v)
        return a
    return v


class ACCESS_DECISION:
    IMPLICIT_DENY = 0
    EXPLICIT_DENY = 1
    EXPLICIT_ALLOW = 2



"""
This library is a linter for AWS IAM policies.
"""
__version__ = "1.5.1"

import fnmatch
import functools
import json
import jsoncfg
import re

import pkg_resources
import yaml

# On initialization, load the IAM data
iam_definition_path = pkg_resources.resource_filename(__name__, "iam_definition.json")
iam_definition = json.load(open(iam_definition_path, "r"))

# And the config data
config_path = pkg_resources.resource_filename(__name__, "config.yaml")
config = yaml.safe_load(open(config_path, "r"))


def override_config(override_config_path):
    if override_config_path is None:
        return

    # Load the override file
    override_config = yaml.safe_load(open(override_config_path, "r"))

    # Over-write the settings
    for finding_type, settings in override_config.items():
        if finding_type not in config:
            config[finding_type] = {}
        for setting, setting_value in settings.items():
            config[finding_type][setting] = setting_value


def enhance_finding(finding):
    if finding.issue not in config:
        raise Exception("Uknown finding issue: {}".format(finding.issue))
    config_settings = config[finding.issue]
    finding.severity = config_settings["severity"]
    finding.title = config_settings["title"]
    finding.description = config_settings.get("description", "")
    finding.ignore_locations = config_settings.get("ignore_locations", None)
    return finding


def analyze_policy_string(
    policy_str,
    filepath=None,
    ignore_private_auditors=False,
    private_auditors_custom_path=None,
    include_community_auditors=False,
    config=None,
):
    """Given a string reperesenting a policy, convert it to a Policy object with findings"""

    try:
        # TODO Need to write my own json parser so I can track line numbers. See https://stackoverflow.com/questions/7225056/python-json-decoding-library-which-can-associate-decoded-items-with-original-li
        policy_json = jsoncfg.loads_config(policy_str)
    except jsoncfg.parser.JSONConfigParserException as e:
        policy = Policy(None)
        policy.add_finding("MALFORMED_JSON", detail="json parsing error: {}".format(e), location={'line': e.line, 'column': e.column})
        return policy

    policy = Policy(policy_json, filepath, config)
    policy.analyze(
        ignore_private_auditors,
        private_auditors_custom_path,
        include_community_auditors,
    )
    return policy


class UnknownPrefixException(Exception):
    pass


class UnknownActionException(Exception):
    pass


def is_arn_match(resource_type, arn_format, resource):
    """
    Match the arn_format specified in the docs, with the resource
    given in the IAM policy.  These can each be strings with globbing. For example, we
    want to match the following two strings:
    - arn:*:s3:::*/*
    - arn:aws:s3:::*personalize*

    That should return true because you could have "arn:aws:s3:::personalize/" which matches both.

    Input:
    - resource_type: Example "bucket", this is only used to identify special cases.
    - arn_format: ARN regex from the docs
    - resource: ARN regex from IAM policy

    
    We can cheat some because after the first sections of the arn match, meaning until the 5th colon (with some
    rules there to allow empty or asterisk sections), we only need to match the ID part.
    So the above is simplified to "*/*" and "*personalize*".

    Let's look at some examples and if these should be marked as a match:
    "*/*" and "*personalize*" -> True
    "*", "mybucket" -> True
    "mybucket", "*" -> True
    "*/*", "mybucket" -> False
    "*/*", "mybucket*" -> True
    "*mybucket", "*myotherthing" -> False
    """
    if arn_format == "*" or resource == "*":
        return True

    if "bucket" in resource_type:
        # We have to do a special case here for S3 buckets
        # and since resources can use variables which contain / need to replace them
        if "/" in strip_var_from_arn(resource, "theVar"):
            return False

    # The ARN has at least 6 parts, separated by a colon. Ensure these exist.
    arn_parts = arn_format.split(":")
    if len(arn_parts) < 6:
        raise Exception("Unexpected format for ARN: {}".format(arn_format))
    resource_parts = resource.split(":")
    if len(resource_parts) < 6:
        raise Exception("Unexpected format for resource: {}".format(resource))

    # For the first 5 parts (ex. arn:aws:SERVICE:REGION:ACCOUNT:), ensure these match appropriately
    # We do this because we don't want "arn:*:s3:::*/*" and "arn:aws:logs:*:*:/aws/cloudfront/*" to return True
    for position in range(0, 5):
        if arn_parts[position] == "*" and resource_parts[position] != "":
            continue
        elif resource_parts[position] == "*":
            continue
        elif arn_parts[position] == resource_parts[position]:
            continue
        else:
            return False

    # Everything up to and including the account ID section matches, so now try to match the remainder
    arn_id = ":".join(arn_parts[5:])
    resource_id = ":".join(resource_parts[5:])

    # Some of the arn_id's contain regexes of the form "[key]" so replace those with "*"
    resource_id = re.sub(r"\[.+?\]", "*", resource_id)
    return is_glob_match(arn_id, resource_id)

def is_arn_strictly_valid(resource_type, arn_format, resource):
    """
    Strictly validate the arn_format specified in the docs, with the resource
    given in the IAM policy. These can each be strings with globbing. For example, we
    want to match the following two strings:
    - arn:*:s3:::*/*
    - arn:aws:s3:::*personalize*

    That should return true because you could have "arn:aws:s3:::personalize/" which matches both.

    However when not using *, must include the resource type in the resource arn and wildcards
    are not valid for the resource type portion (https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-aws-service-namesspaces)

    Input:
    - resource_type: Example "bucket", this is only used to identify special cases.
    - arn_format: ARN regex from the docs
    - resource: ARN regex from IAM policy

    """
    if is_arn_match(resource_type, arn_format, resource):
        # this would have already raised exception
        arn_parts = arn_format.split(":")
        resource_parts = resource.split(":")
        arn_id = ":".join(arn_parts[5:])
        resource_id = ":".join(resource_parts[5:])

        # Does the resource contain a resource type component
        # regex looks for a resource type word like "user" or "cluster-endpoint" followed by a
        # : or / and then anything else excluding the resource type string starting with a *
        arn_id_resource_type = re.match(r"(^[^\*][\w-]+)[\/\:].+", arn_id)

        if arn_id_resource_type != None and resource_id != "*":
            
            # https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-aws-service-namesspaces
            # The following is not allowed: arn:aws:iam::123456789012:u*
            if not (resource_id.startswith(arn_id_resource_type[1])):
                return False

        # replace aws variable and check for other colons
        resource_id_no_vars = strip_var_from_arn(resource_id)
        if ":" in resource_id_no_vars and not ":" in arn_id:
            return False

        return True
    return False

def strip_var_from_arn(arn, replace_with=""):
    return re.sub(r"\$\{aws.[\w\/]+\}", replace_with, arn)

def is_glob_match(s1, s2):
    # This comes from https://github.com/duo-labs/parliament/issues/36#issuecomment-574001764

    # If strings are equal, TRUE
    if s1 == s2:
        return True
    # If either string is all '*'s, TRUE
    if s1 and all(c == "*" for c in s1) or s2 and all(c == "*" for c in s2):
        return True
    # If either string is '', FALSE (already handled case if both are '' in A)
    if not s1 or not s2:
        return False
    # At this point, we know that both s1 and s2 are non-empty, so safe to access [0]'th element
    # If both strings start with '*', TRUE if match first with remainder of second or second with remainder of first
    if s1[0] == s2[0] == "*":
        return is_glob_match(s1[1:], s2) or is_glob_match(s1, s2[1:])
    # If s1 starts with '*', TRUE if remainder of s1 matches any length remainder of s2
    if s1[0] == "*":
        return any(is_glob_match(s1[1:], s2[i:]) for i in range(len(s2)))
    # If s2 starts with '*', TRUE if remainder of s2 matches any length remainder of s1
    if s2[0] == "*":
        return any(is_glob_match(s1[i:], s2[1:]) for i in range(len(s1)))
    # TRUE if s1 and s2 both have same first element and remainder of s1 matches remainder of s2
    return s1[0] == s2[0] and is_glob_match(s1[1:], s2[1:])


@functools.lru_cache(maxsize=1024)
def expand_action(action, raise_exceptions=True):
    """
    Converts "iam:*List*" to
    [
      {'service':'iam', 'action': 'ListAccessKeys'},
      {'service':'iam', 'action': 'ListUsers'}, ...
    ]
    """
    if action == "*":
        action = "*:*"

    parts = action.split(":")
    if len(parts) != 2:
        raise ValueError("Action should be in form service:action")
    prefix = parts[0]
    unexpanded_action = parts[1]

    actions = []
    service_match = None
    for service in iam_definition:
        if service["prefix"] == prefix.lower() or prefix == "*":
            service_match = service

            if len(service["privileges"]) == 0 and prefix != "*":
                # Service has no privileges, so the action must be *
                # For example iq:*
                if unexpanded_action.lower() == "*":
                    return []

            for privilege in service["privileges"]:
                if fnmatch.fnmatchcase(
                    privilege["privilege"].lower(), unexpanded_action.lower()
                ):
                    actions.append(
                        {
                            "service": service_match["prefix"],
                            "action": privilege["privilege"],
                        }
                    )

    if not service_match and raise_exceptions:
        raise UnknownPrefixException("Unknown prefix {}".format(prefix))

    if len(actions) == 0 and raise_exceptions:
        raise UnknownActionException(
            "Unknown action {}:{}".format(prefix, unexpanded_action)
        )

    return actions


def get_resource_type_matches_from_arn(arn):
    """ Given an ARN such as "arn:aws:s3:::mybucket", find resource types that match it.
        This would return:
        [
            "resource": {
                "arn": "arn:${Partition}:s3:::${BucketName}",
                "condition_keys": [],
                "resource": "bucket"
            },
            "service": {
                "service_name": "Amazon S3",
                "privileges": [...]
                ...
            }
        ]
    """
    matches = []
    for service in iam_definition:
        for resource in service["resources"]:
            arn_format = re.sub(r"\$\{.*?\}", "*", resource["arn"])
            if is_arn_match(resource["resource"], arn, arn_format):
                matches.append({"resource": resource, "service": service})
    return matches


def get_privilege_matches_for_resource_type(resource_type_matches):
    """ Given the response from get_resource_type_matches_from_arn(...), this will identify the relevant privileges.
    """
    privilege_matches = []
    for match in resource_type_matches:
        for privilege in match["service"]["privileges"]:
            for resource_type_dict in privilege["resource_types"]:
                resource_type = resource_type_dict["resource_type"].replace("*", "")
                if resource_type == match["resource"]["resource"]:
                    privilege_matches.append(
                        {
                            "privilege_prefix": match["service"]["prefix"],
                            "privilege_name": privilege["privilege"],
                            "resource_type": resource_type,
                        }
                    )

    return privilege_matches


# Import moved here to deal with cyclic dependency
from .policy import Policy



"""Setup script for parliament"""
import os
import re

from setuptools import find_packages, setup


HERE = os.path.dirname(__file__)
VERSION_RE = re.compile(r"""__version__ = ['"]([0-9.]+)['"]""")
TESTS_REQUIRE = ["coverage", "nose"]


def get_version():
    init = open(os.path.join(HERE, "parliament", "__init__.py")).read()
    return VERSION_RE.search(init).group(1)


def get_description():
    return open(
        os.path.join(os.path.abspath(HERE), "README.md"), encoding="utf-8"
    ).read()


setup(
    name="parliament",
    version=get_version(),
    author="Duo Security",
    author_email="scott@summitroute.com",
    description=("parliament audits your AWS IAM policies"),
    long_description=get_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/duo-labs/parliament",
    entry_points={"console_scripts": "parliament=parliament.cli:main"},
    test_suite="tests/unit",
    tests_require=TESTS_REQUIRE,
    extras_require={"dev": TESTS_REQUIRE + ["autoflake", "autopep8", "pylint"]},
    install_requires=["boto3", "jmespath", "pyyaml", "json-cfg"],
    setup_requires=["nose"],
    packages=find_packages(exclude=["tests*"]),
    package_data={
        "parliament": ["iam_definition.json", "config.yaml"],
        "parliament.community_auditors": ["config_override.yaml"],
    },
    zip_safe=True,
    license="BSD 3",
    keywords="aws parliament iam lint audit",
    python_requires=">=3.6",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Development Status :: 5 - Production/Stable",
    ],
)



def audit(policy):
    actions = policy.get_allowed_actions()
    permissions_on_other_users(policy, actions)


# Categories based on https://know.bishopfox.com/blog/5-privesc-attack-vectors-in-aws


def permissions_on_other_users(policy, expanded_actions):
    # Turn into lowercase
    expanded_actions_normalized = [x.lower() for x in expanded_actions]
    expanded_actions = set(expanded_actions_normalized)

    escalation_methods = {
        # 1. IAM Permissions on Other Users
        "CreateAccessKey": ["iam:createaccesskey"],
        "CreateLoginProfile": ["iam:createloginprofile"],
        "UpdateLoginProfile": ["iam:updateloginprofile"],
        # 2. Permissions on Policies
        "CreateNewPolicyVersion": ["iam:createpolicyversion"],
        "SetExistingDefaultPolicyVersion": ["iam:setdefaultpolicyversion"],
        "AttachUserPolicy": ["iam:attachuserpolicy"],
        "AttachGroupPolicy": ["iam:attachgrouppolicy"],
        "AttachRolePolicy": ["iam:attachrolepolicy", "sts:assumerole"],
        "PutUserPolicy": ["iam:putuserpolicy"],
        "PutGroupPolicy": ["iam:putgrouppolicy"],
        "PutRolePolicy": ["iam:putrolepolicy", "sts:assumerole"],
        "AddUserToGroup": ["iam:addusertogroup"],
        # 3. Updating an AssumeRolePolicy
        "UpdateRolePolicyToAssumeIt": ["iam:updateassumerolepolicy", "sts:assumerole"],
        # 4. iam:PassRole:*
        "CreateEC2WithExistingIP": ["iam:passrole", "ec2:runinstances"],
        "PassExistingRoleToNewLambdaThenInvoke": [
            "iam:passrole",
            "lambda:createfunction",
            "lambda:invokefunction",
        ],
        "PassExistingRoleToNewLambdaThenTriggerWithNewDynamo": [
            "iam:passrole",
            "lambda:createfunction",
            "lambda:createeventsourcemapping",
            "dynamodb:createtable",
            "dynamodb:putitem",
        ],
        "PassExistingRoleToNewLambdaThenTriggerWithExistingDynamo": [
            "iam:passrole",
            "lambda:createfunction",
            "lambda:createeventsourcemapping",
        ],
        "PassExistingRoleToNewGlueDevEndpoint": [
            "iam:passrole",
            "glue:createdevendpoint",
        ],
        "PassExistingRoleToCloudFormation": [
            "iam:passrole",
            "cloudformation:createstack",
        ],
        "PassExistingRoleToNewDataPipeline": [
            "iam:passrole",
            "datapipeline:createpipeline",
        ],
        # 5. Privilege Escalation Using AWS Services
        "UpdateExistingGlueDevEndpoint": ["glue:updatedevendpoint"],
        "EditExistingLambdaFunctionWithRole": ["lambda:updatefunctioncode"],
    }

    for key in escalation_methods:
        if set(escalation_methods[key]).issubset(expanded_actions):
            policy.add_finding(
                "PRIVILEGE_ESCALATION",
                location={"type": key, "actions": escalation_methods[key]},
            )
